@@ -10,8 +10,9 @@
 
 module Data.Iteratee.Iteratee (
   -- * Types
+  EnumerateeHandler
   -- ** Error handling
-  throwErr
+  ,throwErr
   ,throwRecoverableErr
   ,checkErr
   -- ** Basic Iteratees
@@ -20,12 +21,15 @@ module Data.Iteratee.Iteratee (
   ,isStreamFinished
   -- ** Chunkwise Iteratees
   ,mapChunksM_
+  ,foldChunksM
   ,getChunk
   ,getChunks
   -- ** Nested iteratee combinators
   ,mapChunks
+  ,mapChunksM
   ,convStream
   ,unfoldConvStream
+  ,unfoldConvStreamCheck
   ,joinI
   ,joinIM
   -- * Enumerators
@@ -43,6 +47,9 @@ module Data.Iteratee.Iteratee (
   -- ** Enumerator Combinators
   ,(>>>)
   ,eneeCheckIfDone
+  ,eneeCheckIfDoneHandle
+  ,eneeCheckIfDoneIgnore
+  ,eneeCheckIfDonePass
   ,mergeEnums
   -- ** Enumeratee Combinators
   ,(><>)
@@ -150,6 +157,14 @@ mapChunksM_ f = liftI step
     step s@(EOF _) = idone () s
 {-# INLINE mapChunksM_ #-}
 
+-- | A fold over chunks
+foldChunksM :: (Monad m, Nullable s) => (a -> s -> m a) -> a -> Iteratee s m a
+foldChunksM f = liftI . go
+  where
+    go a (Chunk c) = lift (f a c) >>= liftI . go
+    go a e = idone a e
+{-# INLINE foldChunksM #-}
+
 -- | Get the current chunk from the stream.
 getChunk :: (Monad m, Nullable s, NullPoint s) => Iteratee s m s
 getChunk = liftI step
@@ -203,12 +218,52 @@ eneeCheckIfDone ::
  (Monad m, NullPoint elo) =>
   ((Stream eli -> Iteratee eli m a) -> Iteratee elo m (Iteratee eli m a))
   -> Enumeratee elo eli m a
-eneeCheckIfDone f inner = Iteratee $ \od oc -> 
+eneeCheckIfDone f = eneeCheckIfDonePass f'
+ where
+  f' k Nothing  = f k
+  f' k (Just e) = throwRecoverableErr e (\s -> joinIM $ enumChunk s $ eneeCheckIfDone f (liftI k))
+
+type EnumerateeHandler eli elo m a =
+  (Stream eli -> Iteratee eli m a)
+  -> SomeException
+  -> Iteratee elo m (Iteratee eli m a)
+
+-- | The same as eneeCheckIfDonePass, with one extra argument:
+-- a handler which is used
+-- to process any exceptions in a separate method.
+eneeCheckIfDoneHandle
+  :: (Monad m, NullPoint elo)
+  => EnumerateeHandler eli elo m a
+  -> ((Stream eli -> Iteratee eli m a)
+      -> Maybe SomeException
+      -> Iteratee elo m (Iteratee eli m a)
+     )
+  -> Enumeratee elo eli m a
+eneeCheckIfDoneHandle h f inner = Iteratee $ \od oc ->
   let onDone x s = od (idone x s) (Chunk empty)
-      onCont k Nothing  = runIter (f k) od oc
-      onCont _ (Just e) = runIter (throwErr e) od oc
+      onCont k Nothing  = runIter (f k Nothing) od oc
+      onCont k (Just e) = runIter (h k e)       od oc
   in runIter inner onDone onCont
 
+eneeCheckIfDonePass
+  :: (Monad m, NullPoint elo)
+  => ((Stream eli -> Iteratee eli m a)
+      -> Maybe SomeException
+      -> Iteratee elo m (Iteratee eli m a)
+     )
+  -> Enumeratee elo eli m a
+eneeCheckIfDonePass f = eneeCheckIfDoneHandle (\k e -> f k (Just e)) f
+{-# INLINE eneeCheckIfDonePass #-}
+
+eneeCheckIfDoneIgnore
+  :: (Monad m, NullPoint elo)
+  => ((Stream eli -> Iteratee eli m a)
+      -> Maybe SomeException
+      -> Iteratee elo m (Iteratee eli m a)
+     )
+  -> Enumeratee elo eli m a
+eneeCheckIfDoneIgnore f = eneeCheckIfDoneHandle (\k _ -> f k Nothing) f
+
 -- | Convert one stream into another with the supplied mapping function.
 -- This function operates on whole chunks at a time, contrasting to
 -- @mapStream@ which operates on single elements.
@@ -217,12 +272,23 @@ eneeCheckIfDone f inner = Iteratee $ \od oc ->
 -- > unpacker = mapChunks B.unpack
 -- 
 mapChunks :: (Monad m, NullPoint s) => (s -> s') -> Enumeratee s s' m a
-mapChunks f = eneeCheckIfDone (liftI . step)
+mapChunks f = eneeCheckIfDonePass (icont . step)
  where
-  step k (Chunk xs)     = eneeCheckIfDone (liftI . step) . k . Chunk $ f xs
+  step k (Chunk xs)     = eneeCheckIfDonePass (icont . step) . k . Chunk $ f xs
   step k str@(EOF mErr) = idone (k $ EOF mErr) str
 {-# INLINE mapChunks #-}
 
+-- | Convert a stream of @s@ to a stream of @s'@ using the supplied function.
+mapChunksM
+  :: (Monad m, NullPoint s, Nullable s)
+  => (s -> m s')
+  -> Enumeratee s s' m a
+mapChunksM f = eneeCheckIfDonePass (icont . step)
+ where
+  step k (Chunk xs)     = lift (f xs) >>=
+                          eneeCheckIfDonePass (icont . step) . k . Chunk
+  step k str@(EOF mErr) = idone (k $ EOF mErr) str
+{-# INLINE mapChunksM #-}
 
 -- |Convert one stream into another, not necessarily in lockstep.
 -- The transformer mapStream maps one element of the outer stream
@@ -235,10 +301,11 @@ convStream ::
  (Monad m, Nullable s) =>
   Iteratee s m s'
   -> Enumeratee s s' m a
-convStream fi = eneeCheckIfDone check
+convStream fi = eneeCheckIfDonePass check
   where
-    check k = isStreamFinished >>= maybe (step k) (idone (liftI k) . EOF . Just)
-    step k = fi >>= eneeCheckIfDone check . k . Chunk
+    check k (Just e) = throwRecoverableErr e (const identity) >> check k Nothing
+    check k _ = isStreamFinished >>= maybe (step k) (idone (liftI k) . EOF . Just)
+    step k = fi >>= eneeCheckIfDonePass check . k . Chunk
 
 -- |The most general stream converter.  Given a function to produce iteratee
 -- transformers and an initial state, convert the stream using iteratees
@@ -248,13 +315,35 @@ unfoldConvStream ::
   (acc -> Iteratee s m (acc, s'))
   -> acc
   -> Enumeratee s s' m a
-unfoldConvStream f acc0 = eneeCheckIfDone (check acc0)
+unfoldConvStream f acc0 = eneeCheckIfDonePass (check acc0)
   where
-    check acc k = isStreamFinished >>=
+    check acc k (Just e) = throwRecoverableErr e (const identity) >> check acc k Nothing
+    check acc k _ = isStreamFinished >>=
                     maybe (step acc k) (idone (liftI k) . EOF . Just)
     step acc k = f acc >>= \(acc', s') ->
-                    eneeCheckIfDone (check acc') . k . Chunk $ s'
+                    eneeCheckIfDonePass (check acc') . k . Chunk $ s'
 
+unfoldConvStreamCheck
+  :: (Monad m, Nullable elo, Monoid elo)
+  => (((Stream eli -> Iteratee eli m a)
+        -> Maybe SomeException
+        -> Iteratee elo m (Iteratee eli m a)
+      )
+      -> Enumeratee elo eli m a
+     )
+  -> (acc -> Iteratee elo m (acc, eli))
+  -> acc
+  -> Enumeratee elo eli m a
+unfoldConvStreamCheck checkDone f acc0 = checkDone (check acc0)
+  where
+    check acc k mX = isStreamFinished >>=
+                   maybe (step acc k mX) (idone (icont k mX) . EOF . Just)
+    step acc k Nothing = f acc >>= \(acc', s') ->
+                  (checkDone (check acc') . k $ Chunk s')
+    step acc k (Just ex) = throwRecoverableErr ex $ \str' ->
+      let i = f acc >>= \(acc', s') ->
+                           (checkDone (check acc') . k $ Chunk s')
+      in joinIM $ enumChunk str' i
 
 -- | Collapse a nested iteratee.  The inner iteratee is terminated by @EOF@.
 --   Errors are propagated through the result.
@@ -294,7 +383,7 @@ type Enumerator s m a = Iteratee s m a -> m (Iteratee s m a)
 -- |Applies the iteratee to the given stream.  This wraps 'enumEof',
 -- 'enumErr', and 'enumPure1Chunk', calling the appropriate enumerator
 -- based upon 'Stream'.
-enumChunk :: (Monad m, Monoid s) => Stream s -> Enumerator s m a
+enumChunk :: (Monad m) => Stream s -> Enumerator s m a
 enumChunk (Chunk xs)     = enumPure1Chunk xs
 enumChunk (EOF Nothing)  = enumEof
 enumChunk (EOF (Just e)) = enumErr e
@@ -371,10 +460,9 @@ mergeEnums e1 e2 etee i = e1 $ e2 (joinI . etee $ ilift lift i) >>= run
 -- 
 -- It passes a given list of elements to the iteratee in one chunk
 -- This enumerator does no IO and is useful for testing of base parsing
-enumPure1Chunk :: (Monad m, Monoid s) => s -> Enumerator s m a
-enumPure1Chunk str iter = runIter iter onDone onCont
+enumPure1Chunk :: (Monad m) => s -> Enumerator s m a
+enumPure1Chunk str iter = runIter iter idoneM onCont
   where
-    onDone a str'    = idoneM a (Chunk str `mappend` str')
     onCont k Nothing = return $ k $ Chunk str
     onCont k e       = return $ icont k e
 
@@ -423,22 +511,24 @@ instance IException NotAnException where
 
 -- |Create an enumerator from a callback function with an exception handler.
 -- The exception handler is called if an iteratee reports an exception.
-enumFromCallbackCatch ::
- (IException e, Monad m, NullPoint s) =>
-  (st -> m (Either SomeException ((Bool, st), s)))
+enumFromCallbackCatch
+  :: (IException e, Monad m, NullPoint s)
+  => (st -> m (Either SomeException ((Bool, st), s)))
   -> (e -> m (Maybe EnumException))
   -> st
   -> Enumerator s m a
 enumFromCallbackCatch c handler = loop
   where
     loop st iter = runIter iter idoneM (onCont st)
-    onCont st k Nothing = c st >>=
-        either (return . k . EOF . Just) (uncurry check)
-      where
-        check (b,st') = if b then loop st' . k . Chunk else return . k . Chunk
+    check k (True,  st') = loop st' . k . Chunk
+    check k (False,_st') = return . k . Chunk
+    onCont st k Nothing  = c st >>=
+        either (return . k . EOF . Just) (uncurry (check k))
     onCont st k j@(Just e) = case fromException e of
-      Just e' -> handler e' >>= maybe (loop st . k $ Chunk empty)
-                                 (return . icont k . Just) . fmap toException
+      Just e' -> handler e' >>=
+                   maybe (loop st . k $ Chunk empty)
+                         (return . icont k . Just) . fmap toException
       Nothing -> return (icont k j)
 {-# INLINE enumFromCallbackCatch #-}
 
+
